"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.chooseJsonFile = exports.getEntryFromZip = exports.downloadDataFile = exports.checkDataUrl = exports.runContainer = exports.buildRunCommand = exports.useRepoVersion = exports.ensureRepo = exports.config = void 0;
const util_1 = __importDefault(require("util"));
const axios_1 = __importDefault(require("axios"));
const readline_sync_1 = __importDefault(require("readline-sync"));
const path_1 = __importDefault(require("path"));
const child_process_1 = require("child_process");
const fs_extra_1 = __importDefault(require("fs-extra"));
const temp_1 = __importDefault(require("temp"));
const zlib_1 = require("zlib");
const promises_1 = require("stream/promises");
const yauzl_1 = __importDefault(require("yauzl"));
exports.config = {
    AVAILABLE_SCHEMAS: [
        'allowed-amounts',
        'in-network-rates',
        'provider-reference',
        'table-of-contents'
    ],
    SCHEMA_REPO_URL: 'https://github.com/CMSgov/price-transparency-guide.git',
    SCHEMA_REPO_FOLDER: path_1.default.normalize(path_1.default.join(__dirname, '..', 'schema-repo'))
};
const ONE_MEGABYTE = 1024 * 1024;
const DATA_SIZE_WARNING_THRESHOLD = ONE_MEGABYTE * 1024; // 1 gigabyte
function ensureRepo(repoDirectory) {
    return __awaiter(this, void 0, void 0, function* () {
        // check if the repo exists, and if not, try to clone it
        if (!fs_extra_1.default.existsSync(path_1.default.join(repoDirectory, '.git'))) {
            return util_1.default.promisify(child_process_1.exec)(`git clone ${exports.config.SCHEMA_REPO_URL} "${repoDirectory}"`);
        }
    });
}
exports.ensureRepo = ensureRepo;
function useRepoVersion(schemaVersion, schemaName, strict = false) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield ensureRepo(exports.config.SCHEMA_REPO_FOLDER);
            const tagResult = yield util_1.default.promisify(child_process_1.exec)(`git -C "${exports.config.SCHEMA_REPO_FOLDER}" tag --list --sort=taggerdate`);
            const tags = tagResult.stdout
                .split('\n')
                .map(tag => tag.trim())
                .filter(tag => tag.length > 0);
            if (tags.includes(schemaVersion)) {
                yield util_1.default.promisify(child_process_1.exec)(`git -C "${exports.config.SCHEMA_REPO_FOLDER}" checkout ${schemaVersion}`);
                let schemaContents = fs_extra_1.default.readFileSync(path_1.default.join(exports.config.SCHEMA_REPO_FOLDER, 'schemas', schemaName, `${schemaName}.json`));
                if (strict) {
                    const modifiedSchema = JSON.parse(schemaContents.toString('utf-8'));
                    makeSchemaStrict(modifiedSchema);
                    schemaContents = JSON.stringify(modifiedSchema);
                }
                // write this version of the schema to a temporary file
                temp_1.default.track();
                const schemaDir = temp_1.default.mkdirSync('schemas');
                const schemaFilePath = path_1.default.join(schemaDir, 'schema.json');
                fs_extra_1.default.writeFileSync(schemaFilePath, schemaContents, { encoding: 'utf-8' });
                return schemaFilePath;
            }
            else {
                // we didn't find your tag. maybe you mistyped it, so show the available ones.
                console.log(`Could not find a schema version named "${schemaVersion}". Available versions are:\n${tags.join('\n')}`);
                process.exitCode = 1;
            }
        }
        catch (error) {
            console.log(`Error when accessing schema: ${error}`);
            process.exitCode = 1;
        }
    });
}
exports.useRepoVersion = useRepoVersion;
// note that this only sets additionalProperties to false at the top level, and at the first level of definitions.
// if there are nested definitions, those will not be modified.
function makeSchemaStrict(schema) {
    if (typeof schema === 'object') {
        schema.additionalProperties = false;
        if (schema.definitions != null && typeof schema.definitions === 'object') {
            for (const defKey of Object.keys(schema.definitions)) {
                schema.definitions[defKey].additionalProperties = false;
            }
        }
    }
}
function buildRunCommand(schemaPath, dataPath, outputPath, containerId) {
    // figure out mount for schema file
    const absoluteSchemaPath = path_1.default.resolve(schemaPath);
    const schemaDir = path_1.default.dirname(absoluteSchemaPath);
    const schemaFile = path_1.default.basename(absoluteSchemaPath);
    // figure out mount for data file
    const absoluteDataPath = path_1.default.resolve(dataPath);
    const dataDir = path_1.default.dirname(absoluteDataPath);
    const dataFile = path_1.default.basename(absoluteDataPath);
    // figure out mount for output file, if provided
    let outputDir, outputFile;
    if ((outputPath === null || outputPath === void 0 ? void 0 : outputPath.length) > 0) {
        const absoluteOutputPath = path_1.default.resolve(outputPath);
        outputDir = path_1.default.dirname(absoluteOutputPath);
        outputFile = path_1.default.basename(absoluteOutputPath);
    }
    if (outputDir && outputFile) {
        return `docker run --rm -v "${schemaDir}":/schema/ -v "${dataDir}":/data/ -v "${outputDir}":/output/ ${containerId} "schema/${schemaFile}" "data/${dataFile}" -o "output/${outputFile}"`;
    }
    else {
        return `docker run --rm -v "${schemaDir}":/schema/ -v "${dataDir}":/data/ ${containerId} "schema/${schemaFile}" "data/${dataFile}"`;
    }
}
exports.buildRunCommand = buildRunCommand;
function runContainer(schemaPath, dataPath, outputPath) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const containerId = yield util_1.default
                .promisify(child_process_1.exec)('docker images validator:latest --format "{{.ID}}"')
                .then(result => result.stdout.trim())
                .catch(reason => {
                console.log(reason.stderr);
                return '';
            });
            if (containerId.length > 0) {
                const runCommand = buildRunCommand(schemaPath, dataPath, outputPath, containerId);
                console.log('Running validator container...');
                return util_1.default
                    .promisify(child_process_1.exec)(runCommand)
                    .then(result => {
                    console.log(result.stdout);
                })
                    .catch(reason => {
                    console.log(reason.stdout);
                    console.log(reason.stderr);
                    process.exitCode = 1;
                });
            }
            else {
                console.log('Could not find a validator docker container.');
                process.exitCode = 1;
            }
        }
        catch (error) {
            console.log(`Error when running validator container: ${error}`);
            process.exitCode = 1;
        }
    });
}
exports.runContainer = runContainer;
function checkDataUrl(url) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            const response = yield axios_1.default.head(url);
            if (response.status === 200) {
                let proceedToDownload;
                const contentLength = parseInt(response.headers['content-length']);
                if (isNaN(contentLength)) {
                    proceedToDownload = readline_sync_1.default.keyInYNStrict('Data file size is unknown. Download this file?');
                }
                else if (contentLength > DATA_SIZE_WARNING_THRESHOLD) {
                    proceedToDownload = readline_sync_1.default.keyInYNStrict(`Data file is ${(contentLength / ONE_MEGABYTE).toFixed(2)} MB in size. Download this file?`);
                }
                else {
                    proceedToDownload = true;
                }
                return proceedToDownload;
            }
            else {
                console.log(`Received unsuccessful status code ${response.status} when checking data file URL.`);
                return false;
            }
        }
        catch (e) {
            console.log('Request failed when checking data file URL.');
            console.log(e.message);
            return false;
        }
    });
}
exports.checkDataUrl = checkDataUrl;
function downloadDataFile(url, folder) {
    return __awaiter(this, void 0, void 0, function* () {
        const filenameGuess = 'data.json';
        const dataPath = path_1.default.join(folder, filenameGuess);
        return new Promise((resolve, reject) => {
            console.log('Beginning download...\n');
            (0, axios_1.default)({
                method: 'get',
                url: url,
                responseType: 'stream',
                onDownloadProgress: progressEvent => {
                    if (process.stdout.isTTY) {
                        let progressText;
                        if (progressEvent.progress != null) {
                            progressText = `Downloaded ${Math.floor(progressEvent.progress * 100)}% of file (${progressEvent.loaded} bytes)`;
                        }
                        else {
                            progressText = `Downloaded ${progressEvent.loaded} bytes`;
                        }
                        process.stdout.clearLine(0, () => {
                            process.stdout.cursorTo(0, () => {
                                process.stdout.write(progressText);
                            });
                        });
                    }
                }
            })
                .then(response => {
                var _a;
                const contentType = (_a = response.headers['content-type']) !== null && _a !== void 0 ? _a : 'application/octet-stream';
                if (isZip(contentType, url)) {
                    // zips require additional work to find a JSON file inside
                    const zipPath = path_1.default.join(folder, 'data.zip');
                    const zipOutputStream = fs_extra_1.default.createWriteStream(zipPath);
                    (0, promises_1.pipeline)(response.data, zipOutputStream).then(() => {
                        yauzl_1.default.open(zipPath, { lazyEntries: true, autoClose: false }, (err, zipFile) => {
                            if (err != null) {
                                reject(err);
                            }
                            const jsonEntries = [];
                            zipFile.on('entry', (entry) => {
                                if (entry.fileName.endsWith('.json')) {
                                    jsonEntries.push(entry);
                                }
                                zipFile.readEntry();
                            });
                            zipFile.on('end', () => {
                                console.log('\nDownload complete.');
                                if (jsonEntries.length === 0) {
                                    reject('No JSON file present in zip.');
                                }
                                else {
                                    let chosenEntry;
                                    if (jsonEntries.length === 1) {
                                        chosenEntry = jsonEntries[0];
                                        zipFile.openReadStream(chosenEntry, (err, readStream) => {
                                            const outputStream = fs_extra_1.default.createWriteStream(dataPath);
                                            outputStream.on('finish', () => {
                                                zipFile.close();
                                                resolve(dataPath);
                                            });
                                            outputStream.on('error', () => {
                                                zipFile.close();
                                                reject('Error writing downloaded file.');
                                            });
                                            readStream.pipe(outputStream);
                                        });
                                    }
                                    else {
                                        jsonEntries.sort((a, b) => {
                                            return a.fileName.localeCompare(b.fileName);
                                        });
                                        resolve({ zipFile, jsonEntries, dataPath });
                                    }
                                }
                            });
                            zipFile.readEntry();
                        });
                    });
                }
                else {
                    const outputStream = fs_extra_1.default.createWriteStream(dataPath);
                    outputStream.on('finish', () => {
                        console.log('\nDownload complete.');
                        resolve(dataPath);
                    });
                    outputStream.on('error', () => {
                        reject('Error writing downloaded file.');
                    });
                    if (isGzip(contentType, url)) {
                        (0, promises_1.pipeline)(response.data, (0, zlib_1.createGunzip)(), outputStream);
                    }
                    else {
                        response.data.pipe(outputStream);
                    }
                }
            })
                .catch(reason => {
                reject('Error downloading data file.');
            });
        });
    });
}
exports.downloadDataFile = downloadDataFile;
function getEntryFromZip(zipFile, entry, dataPath) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve, reject) => {
            zipFile.openReadStream(entry, (err, readStream) => {
                if (err) {
                    reject(err);
                }
                else {
                    const outputStream = fs_extra_1.default.createWriteStream(dataPath);
                    outputStream.on('finish', () => {
                        // keep the zipFile open for now, in case we want more entries
                        resolve();
                    });
                    outputStream.on('error', () => {
                        zipFile.close();
                        reject('Error writing chosen file.');
                    });
                    readStream.pipe(outputStream);
                }
            });
        });
    });
}
exports.getEntryFromZip = getEntryFromZip;
function isGzip(contentType, url) {
    return (contentType === 'application/gzip' ||
        contentType === 'application/x-gzip' ||
        (contentType === 'application/octet-stream' && /\.gz(\?|$)/.test(url)));
}
function isZip(contentType, url) {
    return (contentType === 'application/zip' ||
        (contentType === 'application/octet-stream' && /\.zip(\?|$)/.test(url)));
}
function chooseJsonFile(entries) {
    // there might be a lot of entries. show ten per page of results
    console.log(`${entries.length} JSON files found within ZIP archive.`);
    const maxPage = Math.floor((entries.length - 1) / 10);
    let currentPage = 0;
    let chosenIndex;
    showMenuOptions(currentPage, maxPage, entries.slice(currentPage * 10, currentPage * 10 + 10).map(ent => ent.fileName));
    readline_sync_1.default.promptCLLoop((command, ...extraArgs) => {
        if (/^n(ext)?$/i.test(command)) {
            if (currentPage < maxPage) {
                currentPage++;
            }
            else {
                console.log('Already at last page.');
            }
        }
        else if (/^p(revious)?$/i.test(command)) {
            if (currentPage > 0) {
                currentPage--;
            }
            else {
                console.log('Already at first page.');
            }
        }
        else if (/^go?$/i.test(command)) {
            const targetPage = parseInt(extraArgs[0]);
            if (targetPage > 0 && targetPage <= maxPage + 1) {
                currentPage = targetPage - 1;
            }
            else {
                console.log("Can't go to that page.");
            }
        }
        else if (/^\d$/.test(command)) {
            chosenIndex = currentPage * 10 + parseInt(command);
            console.log(`You selected: ${entries[chosenIndex].fileName}`);
            return true;
        }
        else {
            console.log('Unrecognized command.');
        }
        showMenuOptions(currentPage, maxPage, entries.slice(currentPage * 10, (currentPage + 1) * 10).map(ent => ent.fileName));
    });
    return entries[chosenIndex];
}
exports.chooseJsonFile = chooseJsonFile;
function showMenuOptions(currentPage, maxPage, items) {
    console.log(`Showing page ${currentPage + 1} of ${maxPage + 1}`);
    items.forEach((item, idx) => {
        console.log(`(${idx}): ${item}`);
    });
    const commandsToShow = [];
    if (currentPage > 0) {
        commandsToShow.push('(p)revious page');
    }
    if (currentPage < maxPage) {
        commandsToShow.push('(n)ext page');
    }
    if (maxPage > 0) {
        commandsToShow.push('"(g)o X" to jump to a page');
    }
    console.log(commandsToShow.join(' | '));
}
